﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Mono.Cecil;

namespace MonoCecilExperiments
{
    class Program
    {
        /// <summary>
        ///     Replaces all assembly references with a new one.  Use this method if you want to mock out or replace an existing assembly.
        /// </summary>
        /// <param name="module">The module that has the references you want swapped out.</param>
        /// <param name="oldReference">The old assembly reference you want gone.</param>
        /// <param name="newReference">The new assembly reference you want it replaced with</param>
        static void ReplaceAssemblyReference(ModuleDefinition module, AssemblyNameReference oldReference, AssemblyNameReference newReference)
        {
            for(int i = 0; i < module.AssemblyReferences.Count; i++)
            {
                if (module.AssemblyReferences[i].Equals(oldReference))
                {
                    module.AssemblyReferences[i] = newReference;
                }
            }
        }

        static TypeDefinition FindTypeInAssembly(AssemblyDefinition assembly, string typeName)
        {
            foreach (var module in assembly.Modules)
            {
                foreach (var type in module.GetTypes())
                {
                    if (type.FullName == typeName)
                    {
                        return type;
                    }
                }
            }

            throw new KeyNotFoundException(String.Format("Type '{0}' not found.", typeName));
        }

        static MethodDefinition FindMethodInAssembly(AssemblyDefinition assembly, string methodName)
        {
            foreach (var module in assembly.Modules)
            {
                foreach (var type in module.GetTypes())
                {
                    foreach (var method in type.Methods)
                    {
                        Console.WriteLine(method.FullName);
                        if (method.FullName == methodName)
                        {
                            return method;
                        }
                    }
                }
            }
            throw new KeyNotFoundException(String.Format("Method '{0}' not found.", methodName));
        }

        static void Main(string[] args)
        {
            Console.WriteLine("Loading RewriteMe.exe");
            Mono.Cecil.AssemblyDefinition asm = Mono.Cecil.AssemblyDefinition.ReadAssembly(@"..\..\..\RewriteMe\bin\Debug\RewriteMe.exe");
            Console.WriteLine("Copying over InjectMe assembly");
            System.IO.File.Copy(@"..\..\..\InjectMe\bin\Debug\InjectMe.dll", @"..\..\..\RewriteMe\bin\Debug\InjectMe.dll", true);
            Mono.Cecil.AssemblyDefinition injectAsm = Mono.Cecil.AssemblyDefinition.ReadAssembly(@"..\..\..\RewriteMe\bin\Debug\InjectMe.dll");
            
            Console.WriteLine("Assembly Name: {0}", asm.FullName);
            Console.WriteLine("Injecting Assembly Name: {0}", injectAsm.FullName);

            Console.WriteLine("Adding InjectMe into RewriteMe");
            foreach (var module in asm.Modules)
            {
                Console.WriteLine("Module: {0}", module.FullyQualifiedName);

                var changeBaseClass = FindTypeInAssembly(asm, "RewriteMe.ReplaceMyBaseClass");
                var changeBaseClassTo = FindTypeInAssembly(injectAsm, "InjectMe.BaseClassReplaced");

                Console.WriteLine("Changing base class for {0} from {1} to {2}",
                    changeBaseClass.FullName,
                    changeBaseClass.BaseType.FullName,
                    changeBaseClassTo.FullName);

                // This changes the base type, but not the references in the body
                string oldBaseType = changeBaseClass.BaseType.FullName;
                changeBaseClass.BaseType = module.Import(changeBaseClassTo.GetElementType());

                // We now need to iterate through every single method body and replace calls to the base class
                foreach (var method in changeBaseClass.Methods)
                {
                    var il = method.Body.GetILProcessor();
                    foreach (var instruction in il.Body.Instructions)
                    {
                        if (instruction.OpCode == Mono.Cecil.Cil.OpCodes.Call)
                        {
                            MethodDefinition def = (MethodDefinition)instruction.Operand;
                            Console.WriteLine("Call: {0}", def.FullName);

                            if (def.FullName.Contains(oldBaseType))
                            {
                                Console.WriteLine("Replacing this call.");
                                var injectedMethod = FindMethodInAssembly(injectAsm, def.FullName.Replace(oldBaseType, changeBaseClassTo.FullName));
                                instruction.Operand = method.Module.Import(injectedMethod.Resolve());
                            }
                        }
                    }
                }

                // Let's add a new field for the hell of it
                Console.WriteLine("Adding a new System.Int32 field");
                var int32 = module.Import(typeof(System.Int32));
                var modifyClassDefinition = changeBaseClass.Resolve();
                modifyClassDefinition.Fields.Add(new FieldDefinition("InjectedInt32", FieldAttributes.Public, int32));
            }
            
            asm.Write(@"..\..\..\RewriteMe\bin\Debug\Rewritten.exe");
            Console.WriteLine("Rewritten.exe created.");
            Console.ReadLine();
        }
    }
}
